/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_ppp.h,v 1.3 2006/07/25 19:45:13 sangoma Exp $
 */

/*****************************************************************************
 * sdla_ppp.h	Sangoma PPP firmware API definitions.
 *
 * Author:	Alex Feldman 	<al.feldman@sangoma.com>
 *
 * ============================================================================
 * Dec 06, 1999	Alex Feldman 	Updated to FreeBSD 3.2 
 * Apr 11, 1996	Gene Kozin	Initial version.
 *****************************************************************************
 */

#ifndef	_SDLA_PPP_H
#    define	_SDLA_PPP_H

/*----------------------------------------------------------------------------
 * Notes:
 * ------
 * 1. All structures defined in this file are byte-alined.  
 *
 *	Compiler	Platform
 *	--------	--------
 *	GNU C		Linux		
 */

#ifndef	PACKED
#    define	PACKED	__attribute__((packed))
#endif	/* PACKED */

#include <wanpipe_sppp_iface.h>

/* Adapter memory layout and important constants */
#define	PPP508_MB_VECT	0xE000	/* mailbox window vector */
#define	PPP508_MB_OFFS	0		/* mailbox offset */
#define	PPP508_FLG_OFFS	0x1000	/* status flags offset */
#define	PPP508_BUF_OFFS	0x1100	/* buffer info block offset */
#define PPP514_MB_OFFS  0xE000  /* mailbox offset */
#define PPP514_FLG_OFFS 0xF000  /* status flags offset */
#define PPP514_BUF_OFFS 0xF100  /* buffer info block offset */

#define PPP_MAX_DATA	1008	/* command block data buffer length */

/****** Data Structures *****************************************************/

typedef struct {
	unsigned char 	opp_flag	PACKED;
	unsigned char	trace_type	PACKED;
	unsigned short 	trace_length	PACKED;
	unsigned short 	trace_data_ptr	PACKED;
	unsigned short  trace_time_stamp PACKED;
} trace_element_t;

/* 'command' field defines */
#define PPP_READ_CODE_VERSION	0x10	/* configuration commands */
#define PPP_SET_CONFIG		0x05
#define PPP_READ_CONFIG		0x06
#define	PPP_SET_INTR_FLAGS	0x20
#define	PPP_READ_INTR_FLAGS	0x21
#define	PPP_SET_INBOUND_AUTH	0x30
#define	PPP_SET_OUTBOUND_AUTH	0x31
#define	PPP_GET_CONNECTION_INFO	0x32

#define PPP_COMM_ENABLE		0x03	/* operational commands */
#define PPP_COMM_DISABLE	0x04
#define	PPP_SEND_SIGN_FRAME	0x23
#define	PPP_READ_SIGN_RESPONSE	0x24
#define	PPP_DATALINE_MONITOR	0x33

#define PPP_READ_STATISTICS	0x07	/* statistics commands */
#define PPP_FLUSH_STATISTICS	0x08
#define PPP_READ_ERROR_STATS	0x09
#define PPP_FLUSH_ERROR_STATS	0x0A
#define PPP_READ_PACKET_STATS	0x12
#define PPP_FLUSH_PACKET_STATS	0x13
#define PPP_READ_LCP_STATS	0x14
#define PPP_FLUSH_LCP_STATS	0x15
#define PPP_READ_LPBK_STATS	0x16
#define PPP_FLUSH_LPBK_STATS	0x17
#define PPP_READ_IPCP_STATS	0x18
#define PPP_FLUSH_IPCP_STATS	0x19
#define PPP_READ_IPXCP_STATS	0x1A
#define PPP_FLUSH_IPXCP_STATS	0x1B
#define PPP_READ_PAP_STATS	0x1C
#define PPP_FLUSH_PAP_STATS	0x1D
#define PPP_READ_CHAP_STATS	0x1E
#define PPP_FLUSH_CHAP_STATS	0x1F

#ifdef ENABLE_IPV6
# define PPP_SET_CONFIG_IPV6	0x50
# define PPP_READ_CONFIG_IPV6	0x51
# define	PPP_SET_INBOUND_PAP_AUTH	0x52
# define	PPP_SET_OUTBOUND_PAP_AUTH	0x53
#endif

/* 'result' field defines */
#define PPPRES_OK		0x00	/* command executed successfully */
#define	PPPRES_INVALID_STATE	0x09	/* invalid command in this context */
#define	PPPRES_COMMS_ENABLED	0x09	/* communication is already enabled */

/*----------------------------------------------------------------------------
 * PPP Status Flags.
 *	This structure is located at offset PPP???_FLG_OFFS into
 *	PPP???_MB_VECT.
 */
#define LCP_STATE_OPENED	0x09
#define IP_STATE_OPENED		0x09
typedef struct	ppp_flags
{
	unsigned char iflag		PACKED;	/* 00: interrupt flag */
	unsigned char imask		PACKED;	/* 01: interrupt mask */
	unsigned char resrv		PACKED;
	unsigned char mstatus		PACKED;	/* 03: modem status */
	unsigned char lcp_state		PACKED; /* 04: LCP state */
	unsigned char ppp_phase		PACKED;	/* 05: PPP phase */
	unsigned char ip_state		PACKED; /* 06: IPCP state */
	unsigned char ipx_state		PACKED; /* 07: IPXCP state */
	unsigned char pap_state		PACKED; /* 08: PAP state */
	unsigned char chap_state	PACKED; /* 09: CHAP state */
	unsigned short disc_cause	PACKED;	/* 0A: disconnection cause */
#ifdef ENABLE_IPV6
	unsigned char ipv6_state	PACKED; /* 0C: IPv6CP state */
#endif
} ppp_flags_t;

/* 'iflag' defines */
#define	PPP_INTR_RXRDY		0x01	/* Rx ready */
#define	PPP_INTR_TXRDY		0x02	/* Tx ready */
#define	PPP_INTR_MODEM		0x04	/* modem status change (DCD, CTS) */
#define	PPP_INTR_CMD		0x08	/* interface command completed */
#define	PPP_INTR_DISC		0x10	/* data link disconnected */
#define	PPP_INTR_OPEN		0x20	/* data link open */
#define	PPP_INTR_DROP_DTR	0x40	/* DTR drop timeout expired */
#define PPP_INTR_TIMER          0x80    /* timer interrupt */


/* 'mstatus' defines */
#define	PPP_MDM_DCD		0x08	/* mdm_status: DCD */
#define	PPP_MDM_CTS		0x20	/* mdm_status: CTS */

/* 'disc_cause' defines */
#define PPP_LOCAL_TERMINATION   0x0001	/* Local Request by PPP termination phase */
#define PPP_DCD_CTS_DROP        0x0002  /* DCD and/or CTS dropped. Link down */
#define PPP_REMOTE_TERMINATION	0x0800	/* Remote Request by PPP termination phase */

/* 'misc_config_bits' defines */
#define DONT_RE_TX_ABORTED_I_FRAMES 	0x01
#define TX_FRM_BYTE_COUNT_STATS         0x02
#define RX_FRM_BYTE_COUNT_STATS         0x04
#define TIME_STAMP_IN_RX_FRAMES         0x08
#define NON_STD_ADPTR_FREQ              0x10
#define INTERFACE_LEVEL_RS232           0x20
#define AUTO_LINK_RECOVERY              0x100
#define DONT_TERMINATE_LNK_MAX_CONFIG   0x200                    

/* 'authentication options' defines */
#define NO_AUTHENTICATION	0x00
#define INBOUND_AUTH		0x80
#define PAP_AUTH		0x01
#define CHAP_AUTH		0x02

/* 'ip options' defines */
#define L_AND_R_IP_NO_ASSIG	0x00
#define L_IP_LOCAL_ASSIG    	0x01
#define L_IP_REMOTE_ASSIG   	0x02
#define R_IP_LOCAL_ASSIG        0x04
#define R_IP_REMOTE_ASSIG       0x08
#define ENABLE_IP		0x80

/* 'ipx options' defines */
#define ROUTING_PROT_DEFAULT    0x20
#define ENABLE_IPX		0x80
#define DISABLE_IPX		0x00

#ifdef ENABLE_IPV6
/* 'ipv6 options' defines */
#define IPV6_INTERFACE_IDENTIFIER	0x01
#define IPV6_COMPRESSION_PROT   	0x02
/* FIXME: Temporary value (use ipx_options) */
#define ENABLE_IPv6		0x40 /* 0x80 */
#endif

/*----------------------------------------------------------------------------
 * PPP Buffer Info.
 *	This structure is located at offset PPP508_BUF_OFFS into
 *	PPP508_MB_VECT.
 */
typedef struct	ppp508_buf_info
{
	unsigned short txb_num	PACKED;	/* 00: number of transmit buffers */
	unsigned long  txb_ptr	PACKED;	/* 02: pointer to the buffer ctl. */
	unsigned char  rsrv1[26] PACKED;
	unsigned short rxb_num	PACKED;	/* 20: number of receive buffers */
	unsigned long  rxb_ptr	PACKED;	/* 22: pointer to the buffer ctl. */
	unsigned long  rxb1_ptr	PACKED;	/* 26: pointer to the first buf.ctl. */
	unsigned long  rxb_base	PACKED;	/* 2A: pointer to the buffer base */
	unsigned char  rsrv2[2]	PACKED;
	unsigned long  rxb_end	PACKED;	/* 30: pointer to the buffer end */
} ppp508_buf_info_t;

/* IPV6 Definition for 'proto' member */
#define PROTOCOL_TYPE_IP	0x00
#define PROTOCOL_TYPE_IPX	0x01
#define PROTOCOL_TYPE_RAW	0x02
#define PROTOCOL_TYPE_IPv6	0x03
/*----------------------------------------------------------------------------
 * Transmit/Receive Buffer Control Block.
 */
typedef struct	ppp_buf_ctl
{
	unsigned char  flag		PACKED;	/* 00: 'buffer ready' flag */
	unsigned short length		PACKED;	/* 01: length of data */
	unsigned char  reserved1[1]	PACKED;	/* 03: */
	unsigned char  proto		PACKED;	/* 04: protocol */
	unsigned short timestamp	PACKED;	/* 05: time stamp (Rx only) */
	unsigned char  reserved2[5]	PACKED;	/* 07: */
	union
	{
		unsigned short o_p[2];	/* 1C: buffer offset & page (S502) */
		unsigned long  ptr;	/* 1C: buffer pointer (S508) */
	} buf				PACKED;
} ppp_buf_ctl_t;

/*----------------------------------------------------------------------------
 * S508 Adapter Configuration Block (passed to the PPP_SET_CONFIG command).
 */
typedef struct	ppp508_conf
{
	unsigned long  line_speed	PACKED;	/* 00: baud rate, bps */
	unsigned short txbuf_percent	PACKED;	/* 04: % of Tx buffer */
	unsigned short conf_flags	PACKED;	/* 06: configuration bits */
	unsigned short mtu_local	PACKED;	/* 08: local MTU */
	unsigned short mtu_remote	PACKED;	/* 0A: remote MTU */
	unsigned short restart_tmr	PACKED;	/* 0C: restart timer */
	unsigned short auth_rsrt_tmr	PACKED;	/* 0E: authentication timer */
	unsigned short auth_wait_tmr	PACKED;	/* 10: authentication timer */
	unsigned short mdm_fail_tmr	PACKED;	/* 12: modem failure timer */
	unsigned short dtr_drop_tmr	PACKED;	/* 14: DTR drop timer */
	unsigned short connect_tmout	PACKED;	/* 16: connection timeout */
	unsigned short conf_retry	PACKED;	/* 18: max. retry */
	unsigned short term_retry	PACKED;	/* 1A: max. retry */
	unsigned short fail_retry	PACKED;	/* 1C: max. retry */
	unsigned short auth_retry	PACKED;	/* 1E: max. retry */
	unsigned char  auth_options	PACKED;	/* 20: authentication opt. */
	unsigned char  ip_options	PACKED;	/* 21: IP options */
	unsigned long  ip_local		PACKED;	/* 22: local IP address */
	unsigned long  ip_remote	PACKED;	/* 26: remote IP address */
	unsigned char  ipx_options	PACKED;	/* 2A: IPX options */
	unsigned char  ipx_netno[4]	PACKED;	/* 2B: IPX net number */
	unsigned char  ipx_local[6]	PACKED;	/* 2F: local IPX node number*/
	unsigned char  ipx_remote[6]	PACKED;	/* 35: remote IPX node num.*/
	unsigned char  ipx_router[48]	PACKED;	/* 3B: IPX router name*/
	unsigned long  alt_cpu_clock	PACKED;	/* 6B:  */
} ppp508_conf_t;

/*----------------------------------------------------------------------------
 * S508 Adapter Read Connection Information Block 
 *    Returned by the PPP_GET_CONNECTION_INFO command
 */
typedef struct	ppp508_connect_info
{
	unsigned short 	mru		PACKED;	/* 00-01 Remote Max Rec' Unit */
	unsigned char  	ip_options 	PACKED; /* 02: Negotiated ip options  */
	unsigned long  	ip_local	PACKED;	/* 03-06: local IP address    */
	unsigned long  	ip_remote	PACKED;	/* 07-0A: remote IP address   */
	unsigned char	ipx_options	PACKED; /* 0B: Negotiated ipx options */
	unsigned char  	ipx_netno[4]	PACKED;	/* 0C-0F: IPX net number      */
	unsigned char  	ipx_local[6]	PACKED;	/* 10-1F: local IPX node #    */
	unsigned char  	ipx_remote[6]	PACKED;	/* 16-1B: remote IPX node #   */
	unsigned char  	ipx_router[48]	PACKED;	/* 1C-4B: IPX router name     */
	unsigned char	auth_status	PACKED; /* 4C: Authentication Status  */
	unsigned char 	inbd_auth_peerID[1] PACKED; /* 4D: variable length inbound authenticated peer ID */
} ppp508_connect_info_t;

#ifdef ENABLE_IPV6
/*----------------------------------------------------------------------------
 * PPP IPv6 Adapter Configuration Block (passed to the PPP_SET_CONFIG_IPv6 command).
 */
typedef struct	ppp_ipv6_conf
{
	unsigned char   ipv6_options	PACKED;	/* 0: IPv6 options */
	unsigned char   ipv6_local[16]	PACKED;	/* 1: local IPv6 address */
	unsigned char   ipv6_remote[16]	PACKED;	/* 17: remote IPv6 address */
} ppp_ipv6_conf_t;
#endif

/* 'line_speed' field */
#define	PPP_BITRATE_1200	0x01
#define	PPP_BITRATE_2400	0x02
#define	PPP_BITRATE_4800	0x03
#define	PPP_BITRATE_9600	0x04
#define	PPP_BITRATE_19200	0x05
#define	PPP_BITRATE_38400	0x06
#define	PPP_BITRATE_45000	0x07
#define	PPP_BITRATE_56000	0x08
#define	PPP_BITRATE_64000	0x09
#define	PPP_BITRATE_74000	0x0A
#define	PPP_BITRATE_112000	0x0B
#define	PPP_BITRATE_128000	0x0C
#define	PPP_BITRATE_156000	0x0D

/* Defines for the 'conf_flags' field */
#define	PPP_IGNORE_TX_ABORT	0x01	/* don't re-transmit aborted frames */
#define	PPP_ENABLE_TX_STATS	0x02	/* enable Tx statistics */
#define	PPP_ENABLE_RX_STATS	0x04	/* enable Rx statistics */
#define	PPP_ENABLE_TIMESTAMP	0x08	/* enable timestamp */

/* 'ip_options' defines */
#define	PPP_LOCAL_IP_LOCAL	0x01
#define	PPP_LOCAL_IP_REMOTE	0x02
#define	PPP_REMOTE_IP_LOCAL	0x04
#define	PPP_REMOTE_IP_REMOTE	0x08

/* 'ipx_options' defines */
#define	PPP_REMOTE_IPX_NETNO	0x01
#define	PPP_REMOTE_IPX_LOCAL	0x02
#define	PPP_REMOTE_IPX_REMOTE	0x04
#define	PPP_IPX_ROUTE_RIP_SAP	0x08
#define	PPP_IPX_ROUTE_NLSP	0x10
#define	PPP_IPX_ROUTE_DEFAULT	0x20
#define	PPP_IPX_CONF_COMPLETE	0x40
#define	PPP_IPX_ENABLE		0x80

/*----------------------------------------------------------------------------
 * S508 Adapter Configuration Block (returned by the PPP_READ_CONFIG command).
 */
typedef struct	ppp508_get_conf
{
	unsigned long  bps	PACKED;	/* 00: baud rate, bps */
	ppp508_conf_t  conf	PACKED;	/* 04: requested config. */
	unsigned short txb_num	PACKED;	/* 6F: number of Tx buffers */
	unsigned short rxb_num	PACKED;	/* 71: number of Rx buffers */
} ppp508_get_conf_t;

/*----------------------------------------------------------------------------
 * S508 Operational Statistics (returned by the PPP_READ_STATISTIC command).
 */
typedef struct ppp508_stats
{
	unsigned short reserved1	PACKED;	/* 00: */
	unsigned short rx_bad_len	PACKED;	/* 02: */
	unsigned short reserved2	PACKED;	/* 04: */
	unsigned long  tx_frames	PACKED;	/* 06: */
	unsigned long  tx_bytes		PACKED;	/* 0A: */
	unsigned long  rx_frames	PACKED;	/* 0E: */
	unsigned long  rx_bytes		PACKED;	/* 12: */
} ppp508_stats_t;

/*----------------------------------------------------------------------------
 * Adapter Error Statistics (returned by the PPP_READ_ERROR_STATS command).
 */
typedef struct	ppp_err_stats
{
	unsigned char	 rx_overrun	PACKED;	/* 00: Rx overrun errors */
	unsigned char	 rx_bad_crc	PACKED;	/* 01: Rx CRC errors */
	unsigned char	 rx_abort	PACKED;	/* 02: Rx aborted frames */
	unsigned char	 rx_lost	PACKED;	/* 03: Rx frames lost */
	unsigned char	 tx_abort	PACKED;	/* 04: Tx aborted frames */
	unsigned char	 tx_underrun	PACKED;	/* 05: Tx underrun errors */
	unsigned char	 tx_missed_intr	PACKED;	/* 06: Tx underruns missed */
	unsigned char	 reserved	PACKED;	/* 07: Tx underruns missed */
	unsigned char	 dcd_trans	PACKED;	/* 08: DCD transitions */
	unsigned char	 cts_trans	PACKED;	/* 09: CTS transitions */
} ppp_err_stats_t;



/* Data structure for SET_TRIGGER_INTR command
 */

typedef struct ppp_intr_info{
	unsigned char  i_enable		PACKED; /* 0 Interrupt enable bits */
	unsigned char  irq              PACKED; /* 1 Irq number */
	unsigned short timer_len        PACKED; /* 2 Timer delay */
} ppp_intr_info_t;

#define FT1_MONITOR_STATUS_CTRL                         0x80
#define SET_FT1_MODE                                    0x81

/* Special UDP drivers management commands */
#define PPIPE_ENABLE_TRACING                            0x20
#define PPIPE_DISABLE_TRACING                           0x21
#define PPIPE_GET_TRACE_INFO                            0x22
#define PPIPE_GET_IBA_DATA                              0x23
#define PPIPE_KILL_BOARD     				0x24
#define PPIPE_FT1_READ_STATUS                           0x25
#define PPIPE_DRIVER_STAT_IFSEND                        0x26
#define PPIPE_DRIVER_STAT_INTR                          0x27
#define PPIPE_DRIVER_STAT_GEN                           0x28
#define PPIPE_FLUSH_DRIVER_STATS                        0x29
#define PPIPE_ROUTER_UP_TIME                            0x30

#define DISABLE_TRACING 				0x00
#define TRACE_SIGNALLING_FRAMES				0x01
#define TRACE_DATA_FRAMES				0x02

#define PPP_UDP_SIGNATURE         	"PTPIPEAB"
#define UDP_DRVRSTATS_SIGNATURE     	"DRVSTATS"
#define PPP_UDP_SIGNATURE_LEN		8
  

#ifdef		_MSC_
#  pragma	pack()
#endif
#endif	/* _SDLA_PPP_H */
